/*
 interpreter.c
  Version 1.4d6 (2012.02.10)

  Created: 2008.08.11

 1.4d6
   few fixes to eliminate two warnings.
   _op_srt instead of _op_sqr.
 1.4d5 (2011.09.28)
   expanded parse_textloop to do malloc str, fixed bugs; added \a and \b;
  also added non-standard escapes (whole alphabet).
   freadline no longer uses ungetc.
 1.4d4
   added name hash to wordbreak.
   name & purpose change: parse_line/evaluate_line.
 1.4d3 (2011.08.22)
   added xreadline for easy reading of lines w/ optional escaped-line endings.
 1.4 (2011.06.12)
   added setv for the ability to set identifiers, caller-optional.
  downside: only done by single identifier (no parentheses or ptr setting).
 1.3f2
   parse_number bug fixes: E/e must follow digits, and +/- after digits are not
  part of the notation.
 1.3
   dvs now performs the equiv. of HyperTalk's 'div': full divide, then convert.
   exchanged positions for ceq/cne and cni/cid.
 1.2f8
   standard: fractional digits ignored for integer input instead of syntax error.
   run-on multiplication disallowable when wsp comes in between.
 1.2f7
   optional backward-compatibility: older symbols (CHR_HX).
   fixed an integer-base issue with 'native_eval'.
 1.2f6
   fixed major bugs introduced in 1.2f5; re-added missing CHR_MN macro.
 1.2f5
   'algebraic' changed to 'run-on'.
   op. list and character constants revised.
   CHR_DS (double-space) changed to CHR_NB.
 1.2f4
   floating-point on by default
   can do `.2~/2` (0.2 radical two) and interrogations,
  such as `a?1:2` or `b?4` (zero if b is zero).

 Important Language Notes:
   Unary ops are done *after* run-on multiplication.
    Example: `~/2~/3` does 2*square root of 3, the square root of that (1.8612097182042)
   Unlike C, multiple interrogations (`?`) in a row are not handled, and will error.
    And because the initial 'else' value is taken, `a?(2):1?4:3` will always result
    in 4 (first interrogation, `a?(2):1`, yields non-zero value).

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "interpreter.h"
#include "ftrig.h"

#ifndef ALLOW_RUNON_WSP
 #define ALLOW_RUNON_WSP		0
#endif

#define uint8				unsigned char

#ifdef env68k
 #pragma segment Interpreter

 #define native_double			long double
#else
 #define native_double			double
#endif

#if EVAL_FLOATING
 #define native_eval			native_double
#else
 #define native_eval			evalvalue
#endif

native_double sqrt(native_double);
native_double pow(native_double,
	native_double);
native_double fmod(native_double,
	native_double);
/* scale is `int` under OS X */
native_double scalb(native_double,
#if TARGET_API_MAC_OSX
	int);
#else
	long);
#endif

/* ---- typedefs */

typedef uint8		op_type;

/* ---- constants */

enum
{/* op_type types */
	_op_none,
	_op_expr,	/* 'after value' placeholder; |0x80 = set (`==>`) */
 /* unary */
	_op_ldw,	/* @ */
	_op_srt,	/* ~/	[or]	CHR_RA */

	_op_abs,	/* + */
	_op_neg,	/* -	[or]	CHR_MN */
	_op_not,	/* ~ */
	_op_bno,	/* ! */
 /* dyadic */
	_op_ban,	/* && */
	_op_bor,	/* || */
	_op_and,	/* & */
	_op_ior,	/* | */

	_op_eor,	/* ^ */
	_op_asr,	/* >>> */
	_op_shr,	/* >> */
	_op_shl,	/* << */

	_op_add,	/* + */
	_op_sub,	/* - */
	_op_mul,	/* * */
	_op_pow,	/* ** */

	_op_div,	/* / */
	_op_dvs,	/* //	[or]	CHR_DV */
	_op_mod,	/* % */
	_op_mds,	/* %% */

	_op_clt,	/* < */
	_op_cge,	/* >= */
	_op_cgt,	/* > */
	_op_cle,	/* <= */

	_op_cne,	/* !=	[or]	CHR_NE */
	_op_ceq,	/* = */
	_op_cni,	/* <>	(not identical) */
	_op_cid,	/* ==	(identical) */

	_op_dyad	= _op_ban /* first dyadic -- must update if changed */
};

#ifdef mac
 #define CHR_NE				'\255' /* inequal */
 #define CHR_LE				'\262' /* lesser or equal */
 #define CHR_GE				'\263' /* greater or equal */
 #define CHR_RA				'\303' /* radical (square root) */
 #define CHR_NB				'\312' /* non-breaking space */
 #define CHR_MN				'\320' /* minus "en-dash" */
 #define CHR_DV				'\326' /* divide */
 #define CHR_NL				'\015' /* new line: CR */
#else /* ISO-8895-1-compatible */
/* Windows-1252 */
 //no NE (U+2260)
 //no official LE (U+2264)
 //no official GE (U+2265)
 //no RA (U+221A)
 #define CHR_NB				'\240' /* non-breaking space */
 #define CHR_DV				'\367' /* divide */
 //en-dash (U+2013) not in ISO-8859-1
 #define CHR_MN				'\226' /* minus "dash" */
 #define CHR_NL				'\012' /* new line: LF */
#endif

#ifndef CHR_MN
 #define CHR_MN				'-'
#endif

/* ---- macros */

#define CASE_NUMERIC			case'0':case'1':case'2':case'3':case'4': \
	case'5':case'6':case'7':case'8':case'9':
#define CASE_ALPHABETIC			case'A':case'B':case'C':case'D':case'E': \
	case'F':case'G':case'H':case'I':case'J':case'K':case'L':case'M':case'N': \
	case'O':case'P':case'Q':case'R':case'S':case'T':case'U':case'V':case'W': \
	case'X':case'Y':case'Z':	case'a':case'b':case'c':case'd':case'e': \
	case'f':case'g':case'h':case'i':case'j':case'k':case'l':case'm':case'n': \
	case'o':case'p':case'q':case'r':case's':case't':case'u':case'v':case'w': \
	case'x':case'y':case'z':
#define CASE_BINARY_OPERATOR		case'%':case'&':case'*':case'/':case'<': \
	case'=':case'>':case'^':case'|':

#ifdef CHR_NB
 #define skipwsp(s,e)			do{uint8 c=*(s)++; \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0&&(c+=' '-CHR_NB)!=0) \
		{--(s); break;}}while(s<e)
 #define rskipwsp(s,f)			do{uint8 c=*--(s); \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0&&(c+=' '-CHR_NB)!=0) \
		{++(s); break;}}while(s>f)
#else
 #define skipwsp(s,e)			do{uint8 c=*(s)++; \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0) {--(s); break;}}while(s<e)
 #define rskipwsp(s,f)			do{uint8 c=*--(s); \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0) {++(s); break;}}while(s>f)
#endif

#define skipdig(s,e)			do{uint8 c=*(s)++; \
	if ((c+=-'_')!=0 && (c+='_'-'0')>(uint8)('9'-'0')) \
	{--(s); break;}}while(s<e)

/* ---- globals */

#ifndef NO_TABLES
/* eol 0-3, space/tab/nbsp 9-11, alphanumerics 0x80-0xA3, 0xA4 is underscore */
static const char	upperctype[] =
{
	0x20,0x55,0x0C,0x0D,0x04,0x05,0x06,0x07,0x08,0x09,0x00,0x01,0x02,0x03,0x0E,0x0F,
	0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1A,0x1B,0x1C,0x1D,0x1E,0x1F,
	0x0A,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x29,0x2A,0x2B,0x2C,0x2D,0x2E,0x2F,
	0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x3A,0x3B,0x3C,0x3D,0x3E,0x3F,
	0x40,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,
	0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F,0xA0,0xA1,0xA2,0xA3,0x5B,0x5C,0x5D,0x5E,0xA4,
	0x60,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,
	0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F,0xA0,0xA1,0xA2,0xA3,0x7B,0x7C,0x7D,0x7E,0x7F,
#ifdef mac
	0x30,0x31,0x32,0x33,0x34,0x35,0x36,0xE7,0xCB,0xE5,0x30,0xCC,0x31,0x32,0x33,0xE9,
	0xE6,0xE8,0xEA,0xED,0xEB,0xEC,0x34,0xEE,0xF1,0xEF,0x35,0xCD,0xF2,0xF4,0xF3,0x36,
	0xCA,0x56,0x57,0x58,0x5F,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,
	0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xC6,0xB7,0xB8,0xB8,0xBA,0xBB,0xBC,0xBD,0xAE,0xAF,
	0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0x0B,0xCB,0xCC,0xCD,0xCE,0xCE,
	0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD9,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF,
	0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEA,0xEB,0xEC,0xED,0xEE,0xEF,
	0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF
#else /* windows-1252 */
	0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x41,0x42,0x43,0x44,0x45,0x46,
	0x46,0x47,0x48,0x49,0x4A,0x4B,0x4C,0x4D,0x4E,0x4F,0x50,0x51,0x43,0x52,0x53,0x54,
	0x0B,0x56,0x57,0x58,0x5F,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,
	0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF,
	0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,
	0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF,
	0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,
	0xF0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xF7,0xD8,0xD9,0xDA,0xDB,0xDC,0xFD,0xDE,0x9F
#endif
};

/* ---- system code */

int strcasecmp(const char *a,
	const char *b)
{
	const char	*t = upperctype;
	uint8		c, d;

	if (!a || !b)
		return 1+(-1UL>>1);
	do
		if ((c=t[*((uint8*)a)++])==0)
			return -*b;
	while(c==(d=t[*((uint8*)b)++]));

	return c-d;
}

int strncasecmp(const char *a,
	const char *b,
	size_t z)
{
	const char	*t = upperctype;
	uint8		c, d;

	if (!a || !b)
		return 1+(-1UL>>1);
	if ((signed)z<=0) /* unmapped space? */
		return 0;
	while((c=t[*((uint8*)a)++])==(d=t[*((uint8*)b)++]) && --z)
		;
	return c-d;
}

const char *parse_wordbreak(const char *s,
	const char *e,
	uint8 *h)
{
	if (!s)
		return 0;
	skipwsp(s,e);
	if (s==e)
		return s;
	if (h)
	{
		unsigned	c, v = 0;
		uint8		f = *s&0x3F;

		do
			if (upperctype[c=(uint8)*s++]>(char)0xA4)
				{--s; break;}
		while(v+=c,s<e);
		*h = f|v<<6;
	} else
		do
			if (upperctype[(uint8)*s++]>(char)0xA4)
				{--s; break;}
		while(s<e);
	return s;
}

const char *parse_rwordbreak(const char *s,
	const char *f)
{
	if (!s)
		return 0;
	rskipwsp(s,f);
	if (s==f)
		return s;
	do
		if (upperctype[(uint8)*--s]>(char)0xA4)
			{--s; break;}
	while(s>=f);

	return s;
}
#endif

/* extended standard */

long sreadline(const char *p,
	long z,
	char escape)
{
	const char	*s = p;
	uint8		c, u = 1;

	if (z<=0)
		return 0;
	do
	{
		c = *s++;
		if (escape && c==escape)
			u = 0;
		else
		{
			if ((c-=10)<4) /* LF/VT/FF/CR */
			{
				if (c==3 && *s==10) /* CR+LF = DOS line-ending */
					if (--z)
						++s;
					else
						++z;
				if (u)
					break; /* end of unescaped line */
			}
			u = 1;
		}
	} while(--z);

	return s-p;
}

char *freadline(FILE *f,
	uint32 *length,
	void *(reallocfunc)(void*, size_t),
	uint32 *lecount,
	int escape)
{
	void		*p;
	char		*d = 0;
	long		r, z, lec;
	uint8		e = 0;

	if (!reallocfunc)
		reallocfunc = realloc;
	lec = z = 0;
	if (lecount)
		lec = *lecount;
	if (!length)
		r =-1;
	else if ((r=*length)==0)
		goto end;
	do
	{
		int		c;

		if ((c=fgetc(f))<0) //==EOF)
			break;
		if (escape && c==escape)
			e = 1;
		else
		{
			if ((unsigned)(c-10)<(unsigned)4) /* LF/VT/FF/CR */
			{
				if (e>1)
				{
					e = 0;
					if (c==10) /* (DOS line-ending) */
						continue;
				} else if (c==13)
					e+= e;
				c = '\n', ++lec;
				if (!e) /* end of escaped line */
					break; /* last return not recorded */
			}
			e&=~1;
		}
		if (!(z&7)) /* add some extra bytes for fewer calls */
		{
			if ((p=reallocfunc(d,8+z))==0)
			{
			nix:	d = reallocfunc(d,0);
				goto end;
			}
			d = p;
		}
		d[z] = c;
		if (++z>0x7FFFFFEF)
			goto end; /* can't terminate */
	} while(--r);
	if (d)
	{ /* terminate */
		if (!(z&7) && (p=reallocfunc(d,1+z))==0)
			goto nix;
		d[z] = 0;
	}
end:	if (length)
		*length = z;
	if (lecount)
		*lecount = lec;
	return d;
}

const char *readanyinteger(const char *s,
	long z,
	long int *v)
{
	const char	*e;

	unsigned long int	r = 0, m = 0;

	uint8		c;

	if (!s || z<=0)
		return 0;
	e = z+s, c = *s++;
	if ((c+=-'-')==0)
	{
		if (s>=e)
			return 0;
		m =-1, c = *s++;
	} else if ((c+='-'-'+')==0)
	{
		if (s>=e)
			return 0;
		c = *s++;
	} else if ((uint8)((c+='+')-'0')>(uint8)('9'-'0'))
	{
#ifdef CHR_HX
		if (c==CHR_HX)
			goto hexadecimal;
#endif
		return 0;
	}
	if (c!='0')
	{
		goto decimal;
		do
		{
			c = *s++;
		decimal:
			c-= '0';
			if (c>(uint8)('9'-'0'))
				if ((c+='0'-'A')<=(uint8)('Z'-'A')
						|| (c+='A'-'a')<=(uint8)('z'-'a')
						|| (c+='a'-'_')==0)
					{s = 0; break;}
				else
					{--s; break;}
			r = r*10+c;
		} while(s<e);
		goto end;
	}

	if (s>=e)
		goto end;
	c = *s++;
	if (c!='x' && c!='X')
	{
		if (c=='o' || c=='O')
		{ /* octal (starting-zero-only not supported here) */
			do
			{
				c = *s++, c-= '0';
				if (c>(uint8)('7'-'0'))
					if (c>(uint8)('9'-'7')
							|| (c+='0'-'A')<=(uint8)('Z'-'A')
							|| (c+='A'-'a')<=(uint8)('z'-'a')
							|| (c+='a'-'_')==0)
						{s = 0; break;}
					else
						{--s; break;}
				r = (r<<3)+c;
			} while(s<e);
			goto end;
		}
		if (c=='b' || c=='B')
		{ /* binary */
			do
			{
				c = *s++, c-= '0';
				if (c>(uint8)('1'-'0'))
					if (c>(uint8)('9'-'1')
							|| (c+='0'-'A')<=(uint8)('Z'-'A')
							|| (c+='A'-'a')<=(uint8)('z'-'a')
							|| (c+='a'-'_')==0)
						{s = 0; break;}
					else
						{--s; break;}
				r+= r, r+= c;
			} while(s<e);
			goto end;
		}
		goto decimal;
	}
hexadecimal:
	if (s>=e) /* only starting character? */
		return 0;
	do
	{
		c = *s++, c-= '0';
		if (c<=(uint8)('9'-'0'))
			r = r*16+c;
		else if ((c+='0'-'A')<=(uint8)('Z'-'A')
				|| (c+='A'-'a')<=(uint8)('z'-'a'))
			if (c>(uint8)('F'-'A'))
				{s = 0; break;}
			else
				c+= 10, r = r*16+c;
		else if ((c+='a'-'_')==0)
			{s = 0; break;}
		else
			{--s; break;}
	} while(s<e);
end:	r+= m, r^= m;
	if (v)
		*v = r;

	return s;
}

int makebasestring(void *outstring,
	const char *basechars,
	long value,
	unsigned base,
	int minlength,
	char signchar)
{
	const char	*internalbasechars =
		"0123456789ABCDEFGHIJKLMNOPQRSTUVXYZ@#$`!%&()*+,./:;<=>?[\\]^{|}~";
	uint8		store[32]; /* enough space to do binary */
	uint8		*s = sizeof(store)+store;
	unsigned	v = value;
	int		x = 0;

	if (!base)
		return 0;
	/* SC generated erroneous code if `value` wasn't used in the main loop, */
	/* optimizing it out entirely as `v`, nixing this conditional as a result */
	if ((signed)v<0 && signchar)
		v =-v;
	if (!basechars)
		basechars = internalbasechars;
	if (minlength<=0)
		minlength = 1;
	else if (minlength>sizeof(store))
		minlength = sizeof(store);
	do
	{
		unsigned	i = v;

		*--s = basechars[i-=(v/=base)*base];
	} while(++x<minlength || v);
	minlength = x;

	if (outstring)
	{
		uint8	*d = outstring;

		if (signchar && value<0)
			*d++ = signchar, ++minlength;
		/* copy to output--no func. calls, unaligned compatible */
		if (x&1)
			*d++ = *s++;
		if (x&2)
			*d++ = *s++, *d++ = *s++;
		if (x&=~3)
			do
				*d++ = *s++, *d++ = *s++, *d++ = *s++, *d++ = *s++;
			while(x-=4);
	}

	return minlength;
}

const char *parse_number(const char *s,
	const char *e,
	double *_v,
	unsigned char r)
{
	native_double	fb; //can be a long double, coming from an integer
	double		v = 0;
	int		u = 0, a = 0, b = 10, m = 0, x = 0;
	uint8		c, n = 0;

	if (s>=e)
		return NULL;
	skipwsp(s,e);
#if CHR_MN!='-'
	if ((c=*s++)=='-' || c==(uint8)CHR_MN)
#else
	if ((c=*s++)=='-')
#endif
	{
		n = 1;
		if (s>=e)
			return 0;
		c = *s++;
	} else if (c=='+')
	{
		if (s>=e)
			return 0;
		c = *s++;
	} else if (c!='.' && (uint8)(c-'0')>(uint8)('9'-'0'))
		return 0; /* first must be a digit */

	if (c=='0' && s<e)
	{
		if ((c=*s-'X')==0 || (c+='X'-'x')==0)
			b = 16, u = 3, ++s, r = 0; /* fractionless hexadecimal */
		else if ((c+='x'-'h')==0)
			b = 16, u = 3, ++s; /* hexadecimal w/ fraction */
		else if ((c+='h'-'O')==0 || (c+='O'-'o')==0)
			b = 8, ++s; /* octal (starting-zero-only not supported) */
		else if ((c+='o'-'b')==0)
			b = 2, ++s; /* binary */
		else if ((c+='b'-'d')==0)
			b = 10, r = 0, ++s; /* fractionless decimal */
		else if ((c+='d'-'z')==0)
			b = 36, u = 3, ++s; /* base 36 (all alphanumeric) */
		else
			u|= 0x8000; /* decimal, zero parsed */
		if (s>=e)
			return 0; /* must follow digits */
		c = *s++;
	}
	fb = b;
	goto start;
	do
	{
		c = *s++;
	start:	if ((c+=-'0')<=(uint8)('9'-'0')) /* 0-9 */
		{
			if (c>=(uint8)b)
				return 0; /* digit invalid for this base number */
			u|= 0x8000;
			if (a>0)
				x = 10*x+c;
			else
				v = b*v+c, m+= a;
		} else if ((c+='0'-'A')<='Z'-'A' || (c+='A'-'a')<='z'-'a' && (uint8)u>2)
		{
			if ((c+=10)>=(uint8)b)
				return 0; /* digit invalid for this base number */
			v = b*v+c, m+= a, u|= 0x8000;
			continue;
		} else if ((c+='A'-'E')==0) /* Ee */
		{ /* can't have more than one `E' */
			if (a==1)
				return 0;
			a = 1, u&=~0x8000;
		} else if ((c=s[-1]-'.')==0) /* . */
		{ /* can't have multiple points or for integer or notation */
			if (a || !r)
				return 0;
			a =-1;
#if CHR_MN!='-'
		} else if ((c+='.'-'-')==0 || c==(uint8)(CHR_MN-'-'))
#else
		} else if ((c+='.'-'-')==0)
#endif
		{ /* only one '-' valid here for notation; cannot come after `E+' */
			if (a<=0 || u&0x8001)
				goto back;
			u = 1;
		} else if ((c+='-'-'+')==0) /* + */
		{ /* only one '+' valid here for notation; cannot come after `E-' */
			if (a<=0 || u&0x80FF)
				goto back;
			u = 2;
		} else if ((c+='+'-'_')==0)
			return 0; /* can't directly follow an underscore */
		else /* everything else ends */
		{
		back:	--s;
			break;
		}
	} while(s<e);

	if (!(u&0x8000))
		return 0; /* no digit chars after start or notation */
	if ((uint8)u==1)
		m-= x;
	else
		m+= x;
	if (m)
		v*= pow(b,m);

	if (_v)
		*_v = !n ? v:-v;

	return s;
}

long parse_textloop(const char *t,
	unsigned a,
	char **r,
	void (*b)(unsigned a, unsigned char v))
{
	char		*p = 0;
	const uint8	*s;
	unsigned	m;
	long		z = 0, w;
	uint8		c, q, l;

	if ((s=(const uint8*)t)==0)
		return z;
	z = m = 0, l = 0, w = 3;
	if ((q=*s++)!=0 && q!='"' && q!='\'')
		return ~0; /* was 0 */
	if (r)
		p = *r;
	while(1)
	{
		if ((c=*s++)==0)
			goto bad;
		if (m)
			if (m&0x80000000U)
				goto form;
			else if ((uint8)(c-'0')>(uint8)9)
			{
				if (w>3)
				{
					if ((c+=-'A')<=(uint8)('F'-'A')
							|| (c+='A'-'a')<=(uint8)('f'-'a'))
						{c+= 10, m = m<<4|c; continue;}
					if (m==0x800000) /* no digits */
						goto bad;
				}
			form:	c = m, m = 0, --s;
			} else
				{c&= 7, m = m<<w|c; continue;}
		else if (c==q)
			break;
		else if (c=='\\')
		{
			if ((c=*s++)==0)
				goto bad;
			switch(c)
			{
			case 'h': c = 1; break; /* heading start */
			case 'e': c = 3; break; /* text end/enter */
			case 'q': c = 5; break; /* enquiry */
			case 'k': c = 6; break; /* acknowledge */
			case 'a': c = 7; break; /* bell/alert */
			case 'b': c = 8; break; /* backspace */
			case 't': c = 9; break; /* horiz. tab */
			case 'l': c = 10; break; /* linefeed */
			case 'v': c = 11; break; /* vertical tab */
			case 'f': c = 12; break; /* formfeed */
			case 'r': c = 13; break; /* carriage return */
			case 'o': c = 14; break; /* shift out */
			case 'i': c = 15; break; /* shift in */
			case 'd': c = 16; break; /* data link esc. */
			case 'z': c = 21; break; /* negative ack./zero */
			case 'j': c = 22; break; /* sync. idle/join */
			case 'w': c = 23; break; /* transmission block end/word */
			case 'c': c = 24; break; /* cancel */
			case 'm': c = 25; break; /* medium end */
			case 's': c = 26; break; /* substitute */
			case 'y': c = 27; break; /* escape/yield */
			case 'g': c = 29; break; /* group separator */
			case 'u': c = 31; break; /* unit separator */
			case 'n': c = CHR_NL; break; /* new line */
			case 'p': /* Pascal str */
				if (!z) /* only on starting byte */
					l = 1;
				break;
			case 'x': /* hexadecimal */
				m = 0x800000, w = 4;
				continue;

			default:
				if ((uint8)(c-'0')<=(uint8)7) /* octal */
					{c&= 7, m = c|0x3800000, w = 3; continue;}
				else if ((uint8)(c-10)<=(uint8)3) /* line-ending */
					continue; /* escaped here = ignored */
			}
		}
		if (b)
			b(a++,c);
		if (r)
		{
			if (!(z&7))
				if ((p=realloc(p,8+z))==0)
				{
					*r = realloc(*r,0);
					return 1U<<31;
				} else
					*r = p;
			p[z] = c;
		}
		if (++z>0x7FFFFFF0)
			break;
	}
	if (l)
	{
		if (z>255+1)
		bad:	return ~(s-(const uint8*)t); /* overflow */
		*p = z-1; /* length byte */
	}
	return z;
}

static list_item parse_string(const char *s)
{
	list_item	l;
	char		*r = 0;
	long		z;

	if ((z=parse_textloop(s,0,&r,0))<=0)
		l.pos = 0, l.size =-1<<31, l.data.ptr = 0;
	else
		item_set(&l,0,z,r);

	return l;
}

const char *parse_skipwsp(const char *s,
	const char *e)
{
	if (!s)
		return 0;
	skipwsp(s,e);

	return s;
}

const char *parse_rskipwsp(const char *s,
	const char *f)
{
	if (!s)
		return 0;
	rskipwsp(s,f);

	return s;
}

/* ---- evaluation */

/* single-char op. (e.g., `/`) is unsigned when <=4 bytes */
#if sizeof(evalvalue)>4
 #define evalsing			evalvalue
#else
 #define evalsing			unsigned
#endif

typedef uint8		evalent; /* seperated by dyadics, lower 5 bits are the op. */

/* reads the evalvalue _before_ the pointer */
static evalvalue unpack_eval(const evalent *a)
{
#if sizeof(evalvalue)>4
	uint64		v;

	int64sethl(v,
		((((unsigned)a[-1]<<8)+(unsigned)a[-2]<<8)
		+(unsigned)a[-3]<<8)+(unsigned)a[-4],
		((((unsigned)a[-5]<<8)+(unsigned)a[-6]<<8)
		+(unsigned)a[-7]<<8)+(unsigned)a[-8]);
#else
	uint32		v = *--a;
	int		i = sizeof(evalvalue)-1;

	if (i>0)
		do
			v = (v<<8)|*--a;
		while(--i);
#endif
	return *(const evalvalue*)&v;
}

static evalent *pack_eval(evalent *a,
	evalvalue v)
{
	int		i = sizeof(evalvalue)>4 ? 4:sizeof(evalvalue);
#if sizeof(evalvalue)>4
	const uint64	*p = (const uint64*)&v;
	uint32		n;

	n = int64lo(*p);
	do
		*a++ = n, n>>= 8;
	while(--i);
	n = int64hi(*p), i = 4;
	do
		*a++ = n, n>>= 8;
	while(--i);
#else
	do
		*a++ = v, v>>= 8;
	while(--i);
#endif
	return a;
}

static evalvalue esqrt(evalvalue a)
{
#if EVAL_FLOATING
	return sqrt(a);
#else
	return usqrt(a);
#endif
}

static evalvalue epow(evalvalue a,
	evalvalue b)
{
#if EVAL_FLOATING
	if (a<0)
	{ /* test for odd root of negative (valid) */
		evalvalue	c = 1/b;

		if (!(c-(int)c))
			return -pow(-a,b);
	}
	return pow(a,b);
#else
	if (b<=0)
		return !a ? EVAL_HUGEVAL:!b ? 1:++b ? 0:1/a;
	else if (a==1 || !--b)
		return a;
	else if (b>31)
		return !a ? 0:++a ? EVAL_HUGEVAL:(~(long)b&1)*2-1;
	else
	{
		int64		v;
		evalvalue	m = a;

		do
		{
			(void)muls64s32(v,a,m);
			if ((long)(a=int64lo(v))>>32!=int64hi(v))
				return EVAL_HUGEVAL; /* overflow */
		} while(--b);
	}
	return a;
#endif
}

/*
 Entry bits:
  #7: continue processing (run-on)
  #6: interrogation (`?`), or already packed (doesn't apply here)
  #5: with interrogation, else (`:`)
*/

/* evaluate_chunk evaluates all operations until dyadic w/ bit #7=0;
 returns next stack ptr
*/
static evalent *evaluate_chunk(op_type o,
	const evalent *x,
	evalent *s,
	evalvalue *v,
	evalvalue (*evalindir)(evalvalue))
{
	while((o&=0x9F)!=_op_none)
	{
		native_eval	a, b;

		if (!v)
			goto skip;
		b = *v;
		if (o<_op_dyad)
			a = 0;
		else
			a = unpack_eval(s);
		switch(o&0x1F)
		{ /* unary: op b==>a */
		#include "evaluate.h"

		default: a = b;
		}
		*v = a;
	skip:	if (o>=_op_dyad)
		{
			s-= sizeof(evalvalue);
			if (!(o&0x80))
				break;
		}
		if (s<=x)
			break;
		o = *--s;
	}
end:	return s;
}

/* ---- compilation */

/*
 TO DO:
  offset list for relatives (list of int32 strings)
*/

/* reverse-copy */
#define rcopy8(d,s)			do{uint8 *__d=d; \
	*__d++=*--s,*__d++=*--s,*__d++=*--s,*__d++=*--s; \
	*__d++=*--s,*__d++=*--s,*__d++=*--s,*__d++=*--s; \
}while(0)

static long store_chunk(uint8 **r,
	long z,
	const uint8 *s,
	op_type o)
{
	uint8		*p;

	if (!r)
		return 0;
	p = *r;
	while((o&0x9F)!=_op_none)
	{
		if (!(z&15))
			if ((p=realloc(p,32+z))==0)
			{
				*r = realloc(*r,0);
				return 1U<<31;
			} else
				*r = p;
		if ((p[z]=o)&0x80)
		{ /* immediate */
			rcopy8(p+z+1,s);
			if (*s|s[1] && s[2]|s[3])
				z+= 8;
			else
				*p|= 64, z+= 4;
		}
		++z, o = *--s;
	}
	return z;
}

/* ---- general line processing */

#define EVAL_STACKSIZE			(sizeof(evalvalue)*65)

/* formerly evaluate_parameter */
const char *evaluate_line(const char *s,
	const char *e,
	evalvalue *v,
	const char *(*getv)(const char *s, const char *e, evalvalue*, void *u),
	const char *(*setv)(const char *s, const char *e, evalvalue, void *u),
	void *u,
	evalvalue (*evalindir)(evalvalue))
{
	evalent		xstack[EVAL_STACKSIZE+1]; /* first always _op_none */
	evalent		*a = xstack;
	const char	*p;
	evalparse	i;
	int		n = 0;
	op_type		c; /* operator to be set */
	op_type		o = _op_none; /* previous op. */
	uint8		q = 0; /* interrogation state; 0x40=then, 0x60=else */
#if ALLOW_RUNON_WSP
 #define r		1
#else
	uint8		r = 0; /* when run-on mult. is allowed */
#endif
	do
	{
		switch((uint8)*s) /* try not to get stuck */
		{
		case ' ':
		whitespace:
			skipwsp(s,e);
#if !ALLOW_RUNON_WSP
			r = 0;
#endif
			break;
		case '#':
			++s; /* ignore first '#' for numeric */
		case '.':
			if (s>=e)
				{--s; goto end;} /* fields not implemented */
#ifdef CHR_HX
		case CHR_HX:
#endif
		CASE_NUMERIC
			/* value w/o parentheses */
			if (o==_op_expr || (p=parse_eval(s,e,&i))==0)
				goto end;
			else
				s = p;
			if (!(q&0x40))
				*v = i;
#if !EVAL_FLOATING
			if (*s=='.')
			{ /* skip unused fraction */
				++s;
				skipdig(s,e);
			}
#endif
		runon:	++n; /* increment counter */
#if ALLOW_RUNON_WSP
			skipwsp(s,e);
#endif
			if (s<e)
				if (*s=='(')
					goto paren;
				else if (*s=='@' || *s=='~'
#ifdef CHR_RA
							|| *s==CHR_RA
#endif
						)
				{ /* run-on mul. */
					if (a>=xstack+sizeof(xstack)
							-sizeof(evalvalue)-1)
						return s; /* stack overflow */
					*a++ = o, a = pack_eval(a,*v);
					o = _op_mul|0x80;
					break;
				}
			if ((a=evaluate_chunk(o,xstack,a,q&0x40 ? 0:v,evalindir))<xstack)
				goto end;
			if (s>=e)
				goto end;
			o = _op_expr;
#if !ALLOW_RUNON_WSP
			r = 1;
#endif
			break;
		case '!': c = _op_bno; goto unary;
		case '@': c = _op_ldw; goto unary;
		case '~': c = _op_not;
		unary:	++s;
			skipwsp(s,e);
			if (c==_op_bno && *s=='=') /* != */
				{c = _op_cne; goto dyadic;}
			else if (c==_op_not && *s=='/') /* ~/ */
				c = _op_srt, ++s;
			if (o==_op_expr)
				goto end;
			if (a>=xstack+sizeof(xstack)-1)
				return 0; /* stack overflow */
			else
				*a++ = o;
			o = c;
			break;
		case '+': c = _op_abs;
			if (o==_op_expr)
				{c = _op_add; goto dyadic;}
			goto unary;
		case '-': c = _op_neg;
			if (o==_op_expr)
				{c = _op_sub; goto dyadic;}
			goto unary;
		case '%': c = _op_mod;
			if (o==_op_mod) /* %% */
				c = _op_mds|0x40, o = _op_expr;
			goto dyadic;
		case '&': c = _op_and;
			if (o==_op_and) /* && */
				c = _op_ban|0x40, o = _op_expr;
			goto dyadic;
		case '*': c = _op_mul;
			if (o==_op_mul) /* ** */
				c = _op_pow|0x40, o = _op_expr;
			goto dyadic;
		case '/': c = _op_div;
			if (o==_op_div) /* // */
				c = _op_dvs|0x40, o = _op_expr;
			goto dyadic;
		case '<': c = _op_clt;
			if (o==_op_clt) /* << */
				c = _op_shl|0x40, o = _op_expr;
			goto dyadic;
		case '=': c = _op_ceq;
			if (o==_op_ceq) /* == */
				c = _op_cid|0x40, o = _op_expr;
			else if (o==_op_clt) /* <= */
				c = _op_cle|0x40, o = _op_expr;
			else if (o==_op_cgt) /* >= */
				c = _op_cge|0x40, o = _op_expr;
			goto dyadic;
		case '>': c = _op_cgt;
			if (o==_op_clt) /* <> */
				c = _op_cni|0x40, o = _op_expr;
			else if (o==_op_shr) /* >>> */
				c = _op_asr|0x40, o = _op_expr;
			else if (o==_op_cgt) /* >> */
				c = _op_shr|0x40, o = _op_expr;
			else if (o==_op_cid) /* ==> */
			{
				o = _op_expr|0x80, ++s;
				break;
			}
			goto dyadic;
		case '|': c = _op_ior;
			if (o==_op_ior) /* || */
				c = _op_bor|0x40, o = _op_expr;
			goto dyadic;
		case '^': c = _op_eor;
		dyadic:	if (o!=_op_expr)
			{ /* dyadic after dyadic or unary */
				if (o!=_op_none)
					--s; /* two in a row; discount 2nd */
				goto end;
			}
			if (a>=xstack+sizeof(xstack)-sizeof(evalvalue))
				return 0; /* stack overflow */
			if (!(c&0x40)) /* if not already packed */
				a = pack_eval(a,*v);
			o = c&0x3F, ++s;
			break;
		case ',': /* parameter */
		case ';': /* break */
			goto end;
		case '(':
			if (o==_op_expr)
			{/* run-on multiply */
				if (!r)
					goto end;
			paren:	if (a>=xstack+sizeof(xstack)-sizeof(evalvalue)-2)
					return s; /* stack overflow */
				*a++ = o, a = pack_eval(a,*v), o = _op_mul|0x80;
			} else if (a>=xstack+sizeof(xstack)-1)
				return s; /* stack overflow */
			o&=~0x60, o|= q; /* save interrogation state */
			*a++ = o;
			o = _op_none;
			if ((q&=0x40)==0)
				*v = 0; /* init. interrogation result */
			++s;
			break;
		case ')':
			if (o!=_op_expr || a<=xstack)
				return s; /* after op., empty or underflow */
			o = *--a;
			q = o&0x60; /* restore interrogation state */
			if ((a=evaluate_chunk(o,xstack,a,q&0x40 ? 0:v,evalindir))<xstack)
				goto end;
			o = _op_expr;
#if !ALLOW_RUNON_WSP
			r = 1;
#endif
			++s;
			break;
		case '_':
		CASE_ALPHABETIC
			if (o==(_op_expr|0x80))
			{ /* set value for identifier(s); 2011.06.11 */
				if (!setv)
					goto end; /* no routine */
				if ((p=setv(s,e,unpack_eval(a),u))==0)
					goto end; /* failure to set */
				if ((a-=sizeof(evalvalue))<xstack)
					goto end;
			} else
			{ /* get value for identifier(s) */
				if (!getv || o==_op_expr)
					goto end; /* no routine or two in a row */
				if ((p=getv(s,e,v,u))==0)
					goto end;
				s = p;
				goto runon;
			}
			s = p, ++n; /* update, increment counter */
			o = _op_expr;
			if (s>=e)
				goto end;
			break;
		case '?':
			if (o!=_op_expr || q&0x20)
			{ /* dyadic after dyadic or unary or interrogation open */
				/* no other cases handled here */
				goto end;
			}
			/* eval. 'else' range only if 0; otherwise 'then' only */
			q = *v ? 0x20:0x60;
			*v = 0; /* value parsed; clear */
			o = _op_none;
			++s;
			break;
		case ':':
			if (o!=_op_expr) /* dyadic after dyadic or unary */
				goto end;
			if (!(q&0x20))
			{ /* no other cases handled here */
				goto end;
			}
			q^= 0x60; /* clears bit #5 */
			o = _op_none;
			++s;
			break;
		case 042:
		case 047:
			//(void)parse_string(s);
			break;

		default:
#ifdef CHR_NE
			if ((uint8)*s==(uint8)CHR_NE)
				{c = _op_cne; goto dyadic;}
#endif
#ifdef CHR_LE
			if ((uint8)*s==(uint8)CHR_LE)
				{c = _op_cle; goto dyadic;}
#endif
#ifdef CHR_GE
			if ((uint8)*s==(uint8)CHR_GE)
				{c = _op_cge; goto dyadic;}
#endif
#ifdef CHR_RA
			if ((uint8)*s==(uint8)CHR_RA)
				{c = _op_srt; goto unary;}
#endif
#ifdef CHR_SL
			if ((uint8)*s==(uint8)CHR_SL)
				{c = _op_shl; goto dyadic;}
#endif
#ifdef CHR_SR
			if ((uint8)*s==(uint8)CHR_SR)
				{c = _op_shr; goto dyadic;}
#endif
#if CHR_MN!='-'
			if ((uint8)*s==(uint8)CHR_MN)
				{c = _op_neg; goto unary;}
#endif
#ifdef CHR_DV
			if ((uint8)*s==(uint8)CHR_DV)
				{c = _op_dvs; goto dyadic;}
#endif
#ifdef CHR_NB
			if ((uint8)*s==(uint8)CHR_NB)
				goto whitespace;
#endif
			if ((uint8)*s==(uint8)'\t')
				goto whitespace;
			goto end;
		}
	} while(s<e);
end:	return !n || a>xstack ? 0:s; /* got anything?  incomplete? */
}

static const char *quotedtext_scan(const char *s,
	const char *e,
	uint8 v)
{
	if (e<=s)
		return 0;
	do
	{
		uint8	c = *s++;

		if (c=='\\') /* escape; skip next char */
			++s;
		else if (c==v)
			return s;
	} while(s<e);

	return 0; /* no closing quote */
}

const char *parse_parameter(const char *text,
	long length,
	char *outstring,
	const char *(*getv)(const char *s, const char *e, evalvalue*, void *u),
	const char *(*setv)(const char *s, const char *e, evalvalue, void *u),
	void *ud,
	evalvalue (*evalindir)(evalvalue))
{
	const char	*s, *e;
	evalvalue	v = 0;

	if (outstring)
		*outstring = '\0';

	if ((s=text)==0)
		return 0;
	if (length<=0)
		goto blank;
	e = length+s;
	skipwsp(s,e);

	if (s==e)
	{
	blank:	if (outstring)
			*outstring = ' ';
		goto end;
	}

	if (*s=='\042') /* text - double quote */
		{s = quotedtext_scan(++s,e,'\042'); goto end;}
	else if (*s=='\047') /* text - single quote */
		{s = quotedtext_scan(++s,e,'\047'); goto end;}

	if ((s=evaluate_line(text=s,e,&v,getv,setv,ud,evalindir))==0)
		{s = text; goto end;}
	if (outstring)
#if EVAL_FLOATING
		if (v-(long)v)
			sprintf(outstring,"%.*g",15,v);
		else
#endif
			outstring[makebasestring(outstring,0,v,10,0,'-')] = 0;
end:	return s;
}
